	
###################################################
###	Clubbing 
###################################################


PS.Clubbing	<- function(
  dat,	
	# Note: first column has to contain the year/date.

  adjust.step3 = FALSE, 
	# TRUE means that step (3.3*) is conducted (i.e., the adjusted version of step (3.3)),
	# FALSE means that the unadjusted version step (3.3) is conducted.

  iterate.merging = FALSE,	
	# FALSE means merge only adjacent clubs (i.e. one preliminary club 
	# 	can be merged with one adjacent club (e.g., Phillips & Sul, 2009).
	# TRUE means that club merging can be iterated
	#	(e.g., Panopoulou & Pantelidis, 2009).

  merge.divergencegroup = TRUE,
	# TRUE means, (preliminary) divergence group is allowed to be merged with the previous club 
	#	(if the countries pass the joint log-t-test), which coincides with Phillips & Sul (2009).
	# FALSE means, that countries of a preliminary divergence group can not be merged.

  c.star = 0
	# The critical value for adding single new countries to the core group,
	# compare step (3.2) of algorithm.
){







###
###	Basic computations
###



  t	<- nrow(dat)		# 34 for PS(2009)
  n	<- ncol(dat) - 1	# 152 for PS(2009), -1 because of first column containing the year.





  dat	<- dat[order(dat[, 1], decreasing = FALSE), ]

  y	<- log(dat[, 2:ncol(dat)])
  yh	<- y

  for (i in 1:n){
     comp	<- HPMat2(y[, i], 400)
     yh[,i]	<- comp$gy
  }

  kq	<<- round(t/3)

  period.sort.temp	<- nrow(yh)






###
###	overall-convergence test
###

  (results	<- list(global = logts(yh, kq)))




###
###	Determine first (preliminary) club
###

  stage		<- 1
  outp.temp	<- sorthat(yh = yh, ccc = c.star, period.sort = period.sort.temp, adjust.step3 = adjust.step3)
  prelclub	<- outp.temp$corein2
  remainder	<- outp.temp$cin3
  (results[[paste("stage", stage, sep = "")]]	<- list(
    pcl		= prelclub,
    pcl.logts	= logts(yh[, prelclub], kq),
    rem		= remainder,
    rem.logts	= logts(yh[, remainder], kq)
  ))




###
###	Iterate clubbing to determine all preliminary clubs (i.e. step (4) of algorithm)
###

  repeat{
    stage	<- stage + 1
    outp.temp	<- sorthat(yh = yh[, remainder], ccc = c.star, period.sort = period.sort.temp, adjust.step3 = adjust.step3)
    prelclub	<- remainder[outp.temp$corein2]
    remainder	<- remainder[outp.temp$cin3]
    if(sum(is.na(prelclub)) == 0 & length(remainder) > 0){
      (results[[paste("stage", stage, sep = "")]]	<- list(
        pcl		= prelclub,
        pcl.logts	= logts(yh[, prelclub], kq),
        rem		= remainder,
        rem.logts	= if(length(remainder) > 1){logts(yh[, remainder], kq)} else {NULL}
      ))
      if(length(remainder) < 2){break}
      next
    } else {
      break
      #print("Stop, no core group left.")
    }
  }







###
###	Club merging (i.e. step (5) of algorithm)
###

###	Compute number of stages and check, whether last club is a divergence group

  num.stages	<- max(as.numeric(substr(x = names(results)[grep(names(results), pattern = "stage")], start = 6, stop = 10)))

  pcl.clubs.group	<- rep("club", times = num.stages + 1)
  if(
    any(  
      length(results[[paste("stage", num.stages, sep = "")]]$rem) < 2,
      results[[paste("stage", num.stages, sep = "")]]$rem.logts$tstat_b < -1.65
    )
  ){pcl.clubs.group[num.stages + 1] <- "group"}



###	Create final clubs

  if(merge.divergencegroup){
    pcl.set		<- 1:length(pcl.clubs.group)
  } else {
    pcl.set		<- which(pcl.clubs.group == "club")
  }
  merge.matrix	<- cbind(pcl = pcl.set, fcl = NA)

  pcl.fcl.regions	<- NULL
  for(pcl.temp in pcl.set){
    if(pcl.temp == (num.stages + 1)){
      pcl.fcl.regions	<- rbind(pcl.fcl.regions, cbind(results[[paste("stage", pcl.temp - 1, sep = "")]]$rem, pcl.temp))
    } else {
      pcl.fcl.regions	<- rbind(pcl.fcl.regions, cbind(results[[paste("stage", pcl.temp, sep = "")]]$pcl, pcl.temp))
    }
  }
  colnames(pcl.fcl.regions)	<- c("region", "pcl")






  merge.matrix.b		<- array(data = NA, dim = c(length(pcl.set), length(pcl.set)), dimnames = list(pcl.set, pcl.set))
  merge.matrix.se_b	<- merge.matrix.b
  merge.matrix.tstat_b	<- merge.matrix.b

  for(ro in 1:(length(pcl.set) - 1)){
    if(iterate.merging){
      co.set	<- (ro + 1):max(pcl.set) 
    } else {
      co.set	<- ro + 1
    }
    for(co in co.set){
      logts.temp	<- logts(yh[, pcl.fcl.regions[pcl.fcl.regions[, 2] %in% ro:co, 1]], kq)
      merge.matrix.b[ro, co]		<- logts.temp$b
      merge.matrix.se_b[ro, co]		<- logts.temp$se_b
      merge.matrix.tstat_b[ro, co]	<- logts.temp$tstat_b
    }
  }



  pcl.temp	<- 1
  fcl.temp	<- 1
  pcl.max	<- pcl.temp

  while(any(is.na(merge.matrix[, 2]))){
    if(sum(pcl.fcl.regions[, 2] %in% pcl.temp:pcl.max) > 1){
      while(
        (logts(yh[, pcl.fcl.regions[pcl.fcl.regions[, 2] %in% pcl.temp:pcl.max, 1]], kq)$tstat_b > -1.65) &
        ifelse(test = iterate.merging, yes = (pcl.max <= length(pcl.set)), no = ((pcl.max <= length(pcl.set)) & (pcl.max <= pcl.temp + 1)))
      ){
        merge.matrix[pcl.max, 2]	<- fcl.temp
        pcl.max	<- pcl.max + 1
      }
    } else {
      merge.matrix[pcl.max, 2]	<- fcl.temp
    }
    if((fcl.temp == max(pcl.fcl.regions[, "pcl"])) & (pcl.max <= length(pcl.set))){
      merge.matrix[pcl.max, 2]	<- max(merge.matrix[, "fcl"], na.rm = TRUE) + 1
    } else {
      fcl.temp	<- fcl.temp + 1
    }
    pcl.temp	<- pcl.max
  }













  pcl.fcl.regions		<- merge(x = pcl.fcl.regions, y = merge.matrix, by = "pcl")[, c("region", "pcl", "fcl")]
  if((pcl.clubs.group[length(pcl.clubs.group)] == "group") & (merge.divergencegroup == FALSE)){
    pcl.fcl.divergence	<- data.frame(
      region = (1:n)[!(1:n %in% pcl.fcl.regions[, "region"])], 
      pcl = max(pcl.fcl.regions[, "pcl"]) + 1, 
      fcl = max(pcl.fcl.regions[, "fcl"]) + 1
    )
    pcl.fcl.regions		<- rbind(pcl.fcl.regions, pcl.fcl.divergence)
  }
  pcl.fcl.regions		<- pcl.fcl.regions[order(pcl.fcl.regions[, "region"]), ]




  pcl.divgrp		<- pcl.clubs.group[length(pcl.clubs.group)] == "group"

  fcl.regions.lastgroup	<- pcl.fcl.regions[, "fcl"] == max(pcl.fcl.regions[, "fcl"])
  if(sum(fcl.regions.lastgroup) > 1){
    if(logts(yh[, fcl.regions.lastgroup], kq)$tstat_b > -1.65){
      fcl.divgrp	<- FALSE
    } else {
      fcl.divgrp	<- TRUE
    }
  } else {
    fcl.divgrp	<- TRUE
  }





###	Extend results-list by Club-merging-results

  results[["merge.matrix"]]		<- merge.matrix
  results[["merge.matrix.b"]]		<- merge.matrix.b
  results[["merge.matrix.se_b"]]	<- merge.matrix.se_b
  results[["merge.matrix.tstat_b"]]	<- merge.matrix.tstat_b
  results[["pcl.fcl.regions"]]	<- pcl.fcl.regions
  results[["pcl.divgrp"]]	<- pcl.divgrp
  results[["fcl.divgrp"]]	<- fcl.divgrp

  
  fcl.clubs.group	<- rep("club", times = max(pcl.fcl.regions$fcl))
  if(fcl.divgrp){fcl.clubs.group[length(fcl.clubs.group)]	<- "group"}
  for(cg in 1:length(fcl.clubs.group)){
    results[[paste(fcl.clubs.group[cg], cg, sep = "")]]	<- pcl.fcl.regions[pcl.fcl.regions[, "fcl"] == cg, "region"]
  }


  dat.PS	<<- yh

  return(results)
}










###################################################
###	Functions for computation of PS-clubbing
###################################################

###
###	HPMat2
###

#	Filter time series w.r.t. cyclical and trend components (Hodrick-Prescott)

HPMat2 <- function(	# check
  data,
  lambda
){
  eye		<- diag(length(data))
  gy		<- solve(eye + lambda*crossprod(diff(eye, lag = 1, differences = 2)), data)
  cy		<- data - gy
  result	<- data.frame(gy, cy)
  return(result)
}






###
###	andrs
###

#	Andrews-estimator of long-run variance

andrs		<- function(	# estimate long run variance of errors
  x
){
  t		<- length(x[, 1])
  n		<- length(x[1, ])
  x1		<- x[1:(t - 1), ]
  y1		<- x[2:t, ]
  b1		<- as.matrix(sum(x1*y1)/sum(x1^2))
  ee		<- y1 - x1*b1

  a1		<- (4*b1^2)/(((1 - b1)^2)*((1 + b1)^2))
  a2		<- (4*b1^2)/((1 - b1)^4)
  band1	<- 1.1447*(a1*t)^(1/3)
  band2	<- 1.3221*(a2*t)^(1/5)
  jb2		<- as.matrix((1:(t - 1))/band2)
  jband2	<- jb2*1.2*pi
  kern1	<- ((sin(jband2)/jband2 - cos(jband2))/((jb2*pi)^2*12))*25

  tt		<- length(ee)
  lam		<- as.matrix(0)
  for(j in 1:(tt - 1)){
    ttp1	<- (t(x[1:(tt - j), ]) %*% x[(1 + j):tt, ])*(as.matrix(kern1[j, ]/tt))[1, 1]
    ttp	<- t(t(x[1:(tt - j), ]) %*% x[(1 + j):tt, ])*as.matrix(kern1[j, ]/tt)[1, 1]
    lam	<- lam + ttp + ttp1
  }
  sigm	<- (t(x) %*% x)/as.matrix(tt)
  lam		<- sigm[1] + lam

  return (lam)
}





###
###	logts 
###

#	performs log t-regression and -test

logts	<- function(
  hht, 
  p
){
  t		<- length(hht[, 1])

  trd		<- log(seq(from = 1, by = 1, length.out = t))	# construct log(t)
  trd		<- trd[(p + 1):t]						# also starts at p+1

  hht1	<- as.data.frame(hht)/as.matrix(rowMeans(hht))	# estimate transition coefficients	(h_it)
  hht.temp1	<- (hht1 - 1)^2						# construct cross sectional variance ratio, first (hit-1)^2
  hht.temp2	<- matrix(rowMeans(hht.temp1), ncol = 1)		# second: cross section average 
  lhht	<- log(hht.temp2[1]/hht.temp2)			# taking log(H0/Ht)
  rht		<- lhht[(p + 1):t]					# data begin at t=rT=p+1
  rht		<- rht - 2*log(trd)					# construct log(H0/Ht)-2*log(t)

  xx		<- cbind(1, trd)						# construct design matrix
  b		<- solve(t(xx) %*% xx) %*% t(xx) %*% rht		# OLS
  re		<- rht - xx %*% b						# construct residuals

  lrv		<- andrs(re)    				#long-run variance of errors
  var.b	<- diag(solve(t(xx) %*% xx))*lrv
  se.b	<- sqrt(var.b)
  tstat.b	<- b[2]/se.b[2]
  result	<- data.frame(b = b[2], se_b = se.b[2], tstat_b = tstat.b)
  return(result)
}






###
###	function sorthat
###

#	performs steps (1) to (3) of PS algorithm

sorthat	<- function(
  yh, 
  ccc,
  period.sort, 
  adjust.step3 = FALSE
){

###	Step (1): Sorting
  t		<- length(yh[, 1])
  erg		<- logts(yh, kq)
  n		<- length(yh[1, ])
  stopifnot(period.sort %in% 1:nrow(yh))
  kk		<- cbind(
    t(as.matrix(yh[period.sort, ])), 
    seq(from = 1, by = 1, length.out = n)
  )			# combine country values of period t (i.e. T) and number for each land, last period ordering
  kk		<- kk[rev(order(kk[, 1])), ]		# order land values and number downward with respect to country value (i.e. highest value in first row)
  kin		<- c(t(kk[, 2]))
  yhh		<- yh[, kin]		# order all periods with respect to highest value in last period



###	Step (2.1): Find the two initial core group member countries
  ik1		<- 0
  # search point where H0 is not rejected (where convergence is found)
  erg1	<- matrix(c(0, 0, -2), ncol = 3)
  while(
    (erg1[1, 3] < (-1.65)) && (ik1 < (length(yhh[1,])-1))
  ){
    ik1	<- ik1 + 1
    ww	<- as.matrix(yhh[, ik1:(ik1 + 1)])
    erg1	<- logts(ww, kq)
  }
 

if((erg1$tstat_b > -1.65) & (ncol(yh) > 2)){






###	Step (2.2): Add countries to the two initial core group members
  tt1		<- matrix(c(rep(0, times = n)), ncol = 1)

  i		<- ik1 + 1
  while(i <= length(yhh[1,])){
    ww		<- yhh[, ik1:i]
    erg		<- logts(ww, kq)
    if (erg[1, 3] < (-1.65)) break
    tt1[i, 1]	<- erg[1, 3]
    i			<- i + 1
  }
  kk1		<- which(abs(tt1[, 1]) == max(abs(tt1[, 1])))
  core	<- as.matrix(yhh[, ik1:kk1])

  # core regression ####
  erg		<- logts(core, kq)
  corein1	<- kin[ik1:kk1]



###	Step (3.1): Construct a complement set of core group

  ccin	<- kin[-(ik1:kk1)]
  ccore	<- yhh[, -(ik1:kk1)]



###	Step (3.2): Add country per country to core group (critical value: ccc instead of -1.65).

  corein2	<- corein1
  num.cols	<- if(is.numeric(ccore)){1} else {ncol(ccore)}
  tt2		<- matrix(data = 0, nrow = num.cols, ncol = 1)
  for(i in 1:num.cols){
    ww	<- cbind(core, if(num.cols == 1){ccore} else {ccore[, i]})
    erg	<- logts(ww, kq)
    tt2[i]	<- erg[1, 3]
    if (erg[1, 3] > ccc){ # include region, if t-Stat > critical value ccc
      corein2	<- as.matrix(c(corein2, ccin[i]))
    }
  }
  corein2	<- sort(corein2)


###	Step (3.3)

  cvb		<- logts(yh[, corein2], kq)$tstat_b <= -1.65


  if(cvb){
    if(!adjust.step3){
      corein2	<- corein1	# Manual intervention required, otherwise the initial convergence club is equal to the core group (obtained from step (2))
    } else {

###	Step (3.3*) (i.e., adjusted version of step (3.3))
      K		<- setdiff(corein2, corein1)
      corein2 <- corein1
      while(length(K) > 0){
        erg	<- 1:length(K)
        for (i in 1:length(K)){
          erg[i]	<- logts(cbind(yh[, corein2], yh[, K[i]]), kq)$tstat_b
        }
        cvb2		<- max(erg) <= -1.65
        if(cvb2) break
        corein2	<- c(corein2, K[which(erg == max(erg))])
        K		<- K[-which(erg == max(erg))]
      }
    }
  } 

  cin3	<- setdiff(1:ncol(yh), corein2)

  result	<- list(corein1 = corein1, corein2 = corein2, cin3 = cin3)
} else {result	<- list(corein1 = NA, corein2 = NA, cin3 = 1:n)}
  return(result)
}















